import json
import os
import sys


from pathlib import Path
from subprocess import Popen

from PyQt5.QtWidgets import (
    QApplication,
    QDesktopWidget,
    QFileDialog,
    QGridLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QWidget,
)
from PyQt5.QtGui import QIcon, QIntValidator

from .workflows import (
    prescan_workflow,
    image_workflow,
    data_workflow,
    postprocessing_workflow,
)

PARAMETERS_FILE = "mia-gui-parameters.json"


class MtorImageAnalysis(QWidget):
    def __init__(self):
        super().__init__()
        self.init_ui()

    def init_ui(self):

        # analysis parameters
        self.analysis_parameters = None

        # directory browser
        self.dir_label = QLabel("Image Folder")
        self.dir_btn = QPushButton("Browse")
        self.dir_btn.clicked.connect(self.get_tif_dir)
        self.dir_selected = QLabel()

        # roi input  fields
        self.roi_top_label = QLabel("Top edge of ROI")
        self.roi_top_input = QLineEdit()
        self.roi_top_input.setValidator(QIntValidator())
        self.roi_top_input.textChanged.connect(self.check_button_state)

        self.roi_left_label = QLabel("Left edge of ROI")
        self.roi_left_input = QLineEdit()
        self.roi_left_input.setValidator(QIntValidator())
        self.roi_left_input.textChanged.connect(self.check_button_state)

        self.roi_right_label = QLabel("Right edge of ROI")
        self.roi_right_input = QLineEdit()
        self.roi_right_input.setValidator(QIntValidator())
        self.roi_right_input.textChanged.connect(self.check_button_state)

        self.roi_bottom_label = QLabel("Bottom edge of ROI")
        self.roi_bottom_input = QLineEdit()
        self.roi_bottom_input.setValidator(QIntValidator())
        self.roi_bottom_input.textChanged.connect(self.check_button_state)

        # aditional parameters
        self.pad_x_label = QLabel("Roi padding on x axis")
        self.pad_x_input = QLineEdit()
        self.pad_x_input.setValidator(QIntValidator())
        self.pad_x_input.textChanged.connect(self.check_button_state)

        self.pad_y_label = QLabel("Roi padding on y axis")
        self.pad_y_input = QLineEdit()
        self.pad_y_input.setValidator(QIntValidator())
        self.pad_y_input.textChanged.connect(self.check_button_state)

        self.cc_boost_label = QLabel("Intensity factor for colored images")
        self.cc_boost_input = QLineEdit()
        self.cc_boost_input.setValidator(QIntValidator())
        self.cc_boost_input.textChanged.connect(self.check_button_state)

        # action buttons
        self.btn_run = QPushButton("Run", self)
        self.btn_run.setEnabled(False)
        self.btn_run.clicked.connect(self.run_analysis)

        self.btn_exit = QPushButton("Exit", self)
        self.btn_exit.clicked.connect(QApplication.instance().quit)

        self.fields_map = {
            "folder": (self.dir_selected, str),
            "top": (self.roi_top_input, int),
            "right": (self.roi_right_input, int),
            "bottom": (self.roi_bottom_input, int),
            "left": (self.roi_left_input, int),
            "cut_pad_x": (self.pad_x_input, int),
            "cut_pad_y": (self.pad_y_input, int),
            "boost": (self.cc_boost_input, int),
        }

        # Layout
        grid = QGridLayout()
        grid.setSpacing(10)

        grid.addWidget(self.dir_label, 1, 0)
        grid.addWidget(self.dir_btn, 1, 1)

        grid.addWidget(self.dir_selected, 2, 0, 1, 2)

        grid.addWidget(QLabel(), 3, 0, 1, 2)

        grid.addWidget(self.roi_top_label, 4, 0)
        grid.addWidget(self.roi_top_input, 4, 1)

        grid.addWidget(self.roi_right_label, 5, 0)
        grid.addWidget(self.roi_right_input, 5, 1)

        grid.addWidget(self.roi_bottom_label, 6, 0)
        grid.addWidget(self.roi_bottom_input, 6, 1)

        grid.addWidget(self.roi_left_label, 7, 0)
        grid.addWidget(self.roi_left_input, 7, 1)

        grid.addWidget(QLabel(), 8, 0, 1, 2)

        grid.addWidget(self.pad_x_label, 9, 0)
        grid.addWidget(self.pad_x_input, 9, 1)

        grid.addWidget(self.pad_y_label, 10, 0)
        grid.addWidget(self.pad_y_input, 10, 1)

        grid.addWidget(self.cc_boost_label, 11, 0)
        grid.addWidget(self.cc_boost_input, 11, 1)

        grid.addWidget(self.btn_exit, 12, 0)
        grid.addWidget(self.btn_run, 12, 1)

        # set default values
        self.pad_x_input.setText("50")
        self.pad_y_input.setText("25")
        self.cc_boost_input.setText("5")

        self.setLayout(grid)
        # self.resize(350, 300)
        self.center_window_on_desktop()
        self.setWindowTitle("mTor Image Analysis")
        self.setWindowIcon(QIcon("web.png"))

        self.show()

    def center_window_on_desktop(self):
        cp = QDesktopWidget().availableGeometry().center()
        fg = self.frameGeometry()
        fg.moveCenter(cp)
        self.move(fg.topLeft())

    def get_tif_dir(self):
        dlg = QFileDialog(directory=str(Path.home()))
        dlg.setFileMode(QFileDialog.Directory)

        if dlg.exec_():
            filenames = dlg.selectedFiles()
            folder = Path(filenames[0])
            items = (i for i in folder.iterdir() if not i.stem.startswith("."))
            tifs = [i for i in items if i.suffix == ".tif"]
            if len(tifs) == 0:
                self.dir_selected.setText("")
                self.check_button_state()
            else:
                self.dir_selected.setText(filenames[0])
                parameter_path = folder / PARAMETERS_FILE
                if parameter_path.is_file():
                    try:
                        with parameter_path.open("r") as fh:
                            parameters = json.load(fh)
                            for key, value in parameters.items():
                                field = self.fields_map[key][0]
                                field.setText(str(value))
                    except Exception:
                        pass
                self.check_button_state()

    def check_button_state(self):
        parameters = self.get_parameters()
        self.btn_run.setEnabled(all(parameters.values()))

    def get_parameters(self):
        parameters = {key: None for key in self.fields_map}
        for key, field_and_func in self.fields_map.items():
            field, func = field_and_func
            raw_data = field.text().strip()
            if raw_data:
                if func is int:
                    # in qt5, a point as 1000 separator is allowed
                    raw_data = raw_data.replace(".", "")
                parameters[key] = func(raw_data)
        return parameters

    def run_analysis(self):
        parameters = self.get_parameters()
        if all(parameters.values()):
            parameter_path = Path(parameters["folder"]) / PARAMETERS_FILE
            try:
                with parameter_path.open("w") as fh:
                    tmp = {
                        k: v for k, v in parameters.items() if k != "folder"
                    }
                    json.dump(tmp, fh)
            except Exception:
                pass

            self.analysis_parameters = parameters
            self.hide()
            QApplication.instance().quit()


def run_gui():
    import traceback

    try:

        app = QApplication(sys.argv)
        mia = MtorImageAnalysis()
        app.exec_()

        if mia.analysis_parameters is not None:
            pw = prescan_workflow(**mia.analysis_parameters)
            iw = image_workflow(pw.parameters)
            dw = data_workflow(iw.data, iw.parameters)
            fw = postprocessing_workflow(dw.data, dw.parameters)
            pdf_path = fw.parameters.data_dir / "report.pdf"
            if sys.platform.startswith("win"):
                os.startfile(str(pdf_path))
            else:
                if sys.platform.startswith("darwin"):
                    cmd = ["open", str(pdf_path)]
                else:
                    cmd = ["xdg-open", str(pdf_path)]
                Popen(
                    cmd,
                    shell=False,
                    stdin=None,
                    stdout=None,
                    stderr=None,
                    close_fds=True,
                )
    except Exception as err:
        traceback.print_exc()
        input("Press enter to close this window")
    finally:
        sys.exit()
